/*
 * Copyright 2010 Jerome Glisse <glisse@freedesktop.org>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * on the rights to use, copy, modify, merge, publish, distribute, sub
 * license, and/or sell copies of the Software, and to permit persons to whom
 * the Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHOR(S) AND/OR THEIR SUPPLIERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 * USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * Authors:
 *      Jerome Glisse
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "radeon_pci.h"
#include "atom.h"

struct ctx {
    struct radeon_pci       pradeon;
    void                    *bios;
    unsigned                bios_size;
    struct atom_context     *atom_ctx;
};

/* ATOM accessor methods */
static uint32_t cail_pll_read(struct card_info *info, uint32_t reg)
{
    fprintf(stderr, "%s:%d unsupported operation\n", __func__, __LINE__);\
    return 0;
}

static void cail_pll_write(struct card_info *info, uint32_t reg, uint32_t val)
{
    fprintf(stderr, "%s:%d unsupported operation\n", __func__, __LINE__);\
}

static uint32_t cail_mc_read(struct card_info *info, uint32_t reg)
{
    fprintf(stderr, "%s:%d unsupported operation\n", __func__, __LINE__);\
    return 0;
}

static void cail_mc_write(struct card_info *info, uint32_t reg, uint32_t val)
{
    fprintf(stderr, "%s:%d unsupported operation\n", __func__, __LINE__);
}

static void cail_reg_write(struct card_info *info, uint32_t reg, uint32_t val)
{
    struct ctx *ctx = info->ctx;

    radeon_mmio_wr32(&ctx->pradeon, reg << 2, val);
}

static uint32_t cail_reg_read(struct card_info *info, uint32_t reg)
{
    struct ctx *ctx = info->ctx;

    return radeon_mmio_rd32(&ctx->pradeon, reg << 2);
}

#define R600_BIOS_0_SCRATCH               0x1724
#define R600_BIOS_1_SCRATCH               0x1728
#define R600_BIOS_2_SCRATCH               0x172c
#define R600_BIOS_3_SCRATCH               0x1730
#define R600_BIOS_4_SCRATCH               0x1734
#define R600_BIOS_5_SCRATCH               0x1738
#define R600_BIOS_6_SCRATCH               0x173c
#define R600_BIOS_7_SCRATCH               0x1740

void atom_initialize_bios_scratch_regs(struct ctx *ctx)
{
    uint32_t bios_2_scratch, bios_6_scratch;

    bios_2_scratch = radeon_mmio_rd32(&ctx->pradeon, R600_BIOS_2_SCRATCH);
    bios_6_scratch = radeon_mmio_rd32(&ctx->pradeon, R600_BIOS_6_SCRATCH);

    /* let the bios control the backlight */
    bios_2_scratch &= ~ATOM_S2_VRI_BRIGHT_ENABLE;

    /* tell the bios not to handle mode switching */
    bios_6_scratch |= ATOM_S6_ACC_BLOCK_DISPLAY_SWITCH;
    bios_6_scratch &= ~ATOM_S6_ACC_BLOCK_DISPLAY_SWITCH;

    radeon_mmio_wr32(&ctx->pradeon, R600_BIOS_2_SCRATCH, bios_2_scratch);
    radeon_mmio_wr32(&ctx->pradeon, R600_BIOS_6_SCRATCH, bios_6_scratch);
}

int atombios_init(struct ctx *ctx)
{
    struct card_info atom_card_info = {
        .reg_read       = cail_reg_read,
        .reg_write      = cail_reg_write,
        .ioreg_read     = cail_reg_read,
        .ioreg_write    = cail_reg_write,
        .mc_read        = cail_mc_read,
        .mc_write       = cail_mc_write,
        .pll_read       = cail_pll_read,
        .pll_write      = cail_pll_write,
    };

    atom_card_info.ctx = ctx;
    ctx->atom_ctx = atom_parse(&atom_card_info, ctx->bios);
    if (ctx->atom_ctx == NULL) {
        return -EINVAL;
    }
    atom_initialize_bios_scratch_regs(ctx);
    atom_allocate_fb_scratch(ctx->atom_ctx);
    return 0;
}

void atombios_fini(struct ctx *ctx)
{
    if (ctx->atom_ctx) {
        free(ctx->atom_ctx->scratch);
        free(ctx->atom_ctx);
    }
}

struct log {
    unsigned data[8];
};

static struct log logs[] = {
// 1024x768
{{9, 0x00000002, 0x00010200, 0x000c1e12, 0x00012d78, 0xb1f8220c, 0x00010ff8, 0x01080e4a}},
{{78, 0x22b83e00, 0x01000091, 0x00010003, 0x2deb2e3c, 0xb1f82218, 0x00010ff8, 0x01080e4a}},
{{78, 0x22c03e00, 0x01000091, 0x00010003, 0x22182e0e, 0x0ff8b1f8, 0x0e4a0001, 0x00030108}},
{{50, 0x14003105, 0x00000200, 0x22240001, 0x0ff8b1f8, 0x0e4a0001, 0x00030108, 0x4f3f530c}},
{{78, 0x00003e00, 0x01000091, 0x00010003, 0xb1f82235, 0x00010ff8, 0x01080e4a, 0x530c0003}},
{{78, 0x00003e00, 0x01000091, 0x00010103, 0xb1f82235, 0x00010ff8, 0x01080e4a, 0x530c0003}},
{{78, 0x20203e00, 0x01000091, 0x00010003, 0xb1f82242, 0x00010ff8, 0x01080e4a, 0x530c0003}},
{{78, 0x20203e00, 0x01000091, 0x00010103, 0xb1f82242, 0x00010ff8, 0x01080e4a, 0x530c0003}},
{{78, 0x3e4c3e00, 0x01000091, 0x00010003, 0xb1f82254, 0x00010ff8, 0x01080e4a, 0x530c0003}},
{{78, 0x00003e00, 0x01000091, 0x00010003, 0xb1f82235, 0x00010ff8, 0x01080e4a, 0x530c0003}},
{{78, 0x00003e00, 0x01000091, 0x00010103, 0xb1f82235, 0x00010ff8, 0x01080e4a, 0x530c0003}},
{{78, 0x20303e00, 0x01000091, 0x00010003, 0xb1f82242, 0x00010ff8, 0x01080e4a, 0x530c0003}},
{{78, 0x20303e00, 0x01000091, 0x00010103, 0xb1f82242, 0x00010ff8, 0x01080e4a, 0x530c0003}},
{{78, 0x3e4c3e00, 0x01000091, 0x00010003, 0xb1f82254, 0x00010ff8, 0x01080e4a, 0x530c0003}},
{{78, 0x00003e00, 0x01000091, 0x00010003, 0xb1f82235, 0x00010ff8, 0x01080e4a, 0x530c0003}},
{{78, 0x00003e00, 0x01000091, 0x00010103, 0xb1f82235, 0x00010ff8, 0x01080e4a, 0x530c0003}},
{{78, 0x20403e00, 0x01000091, 0x00010003, 0xb1f82242, 0x00010ff8, 0x01080e4a, 0x530c0003}},
{{78, 0x20403e00, 0x01000091, 0x00010103, 0xb1f82242, 0x00010ff8, 0x01080e4a, 0x530c0003}},
{{78, 0x3e4c3e00, 0x01000091, 0x00010003, 0xb1f82254, 0x00010ff8, 0x01080e4a, 0x530c0003}},
{{78, 0x00003e00, 0x01000091, 0x00010003, 0xb1f82235, 0x00010ff8, 0x01080e4a, 0x530c0003}},
{{78, 0x00003e00, 0x01000091, 0x00010103, 0xb1f82235, 0x00010ff8, 0x01080e4a, 0x530c0003}},
{{78, 0x20503e00, 0x01000091, 0x00010003, 0xb1f82242, 0x00010ff8, 0x01080e4a, 0x530c0003}},
{{78, 0x20503e00, 0x01000091, 0x00010103, 0xb1f82242, 0x00010ff8, 0x01080e4a, 0x530c0003}},
{{78, 0x3e4c3e00, 0x01000091, 0x00010003, 0xb1f82254, 0x00010ff8, 0x01080e4a, 0x530c0003}},
{{78, 0x00003e00, 0x01000091, 0x00010003, 0xb1f82235, 0x00010ff8, 0x01080e4a, 0x530c0003}},
{{78, 0x00003e00, 0x01000091, 0x00010103, 0xb1f82235, 0x00010ff8, 0x01080e4a, 0x530c0003}},
{{78, 0x20603e00, 0x01000091, 0x00010003, 0xb1f82242, 0x00010ff8, 0x01080e4a, 0x530c0003}},
{{78, 0x20603e00, 0x01000091, 0x00010103, 0xb1f82242, 0x00010ff8, 0x01080e4a, 0x530c0003}},
{{78, 0x3e4c3e00, 0x01000091, 0x00010003, 0xb1f82254, 0x00010ff8, 0x01080e4a, 0x530c0003}},
{{78, 0x00003e00, 0x01000091, 0x00010003, 0xb1f82235, 0x00010ff8, 0x01080e4a, 0x530c0003}},
{{78, 0x00003e00, 0x01000091, 0x00010103, 0xb1f82235, 0x00010ff8, 0x01080e4a, 0x530c0003}},
{{78, 0x20703e00, 0x01000091, 0x00010003, 0xb1f82242, 0x00010ff8, 0x01080e4a, 0x530c0003}},
{{78, 0x20703e00, 0x01000091, 0x00010103, 0xb1f82242, 0x00010ff8, 0x01080e4a, 0x530c0003}},
{{78, 0x3e4c3e00, 0x01000091, 0x00010003, 0xb1f82254, 0x00010ff8, 0x01080e4a, 0x530c0003}},
{{78, 0x00003e00, 0x01000091, 0x00010003, 0xb1f82235, 0x00010ff8, 0x01080e4a, 0x530c0003}},
{{78, 0x00003e00, 0x01000091, 0x00010103, 0xb1f82235, 0x00010ff8, 0x01080e4a, 0x530c0003}},
{{78, 0x20803e00, 0x01000091, 0x00010003, 0xb1f82242, 0x00010ff8, 0x01080e4a, 0x530c0003}},
{{78, 0x20803e00, 0x01000091, 0x00010103, 0xb1f82242, 0x00010ff8, 0x01080e4a, 0x530c0003}},
{{78, 0x3e4c3e00, 0x01000091, 0x00010003, 0xb1f82254, 0x00010ff8, 0x01080e4a, 0x530c0003}},
{{78, 0x00003e00, 0x01000091, 0x00010003, 0xb1f82235, 0x00010ff8, 0x01080e4a, 0x530c0003}},
{{78, 0x00003e00, 0x01000091, 0x00010103, 0xb1f82235, 0x00010ff8, 0x01080e4a, 0x530c0003}},
{{78, 0x20903e00, 0x01000091, 0x00010003, 0xb1f82242, 0x00010ff8, 0x01080e4a, 0x530c0003}},
{{78, 0x20903e00, 0x01000091, 0x00010103, 0xb1f82242, 0x00010ff8, 0x01080e4a, 0x530c0003}},
{{78, 0x3e4c3e00, 0x01000091, 0x00010003, 0xb1f82254, 0x00010ff8, 0x01080e4a, 0x530c0003}},
{{54, 0x00322618, 0x80002618, 0x228d90a0, 0x0ff8b1f8, 0x0e4a0008, 0x00030108, 0x4f3f530c}},
{{54, 0x00322618, 0x80002618, 0x229490a2, 0x0ff8b1f8, 0x0e4a0008, 0x00030108, 0x4f3f530c}},
{{4, 0x0c00002f, 0x0000ff03, 0x2ddd0001, 0x16df000c, 0x1ecb168e, 0x4f3f5317, 0x00000000}},
{{34, 0x17dd0100, 0x16b41886, 0x1ecb1699, 0x4f3f5317, 0x00000000, 0x00000000, 0x00000000}},
{{6, 0x17dd0000, 0x16be1886, 0x16990201, 0x53171ecb, 0x00004f3f, 0x00000000, 0x00000000}},
{{44, 0x17dd0100, 0x18941886, 0x53171ed0, 0x00004f3f, 0x00000000, 0x00000000, 0x01230000}},
{{76, 0x00842221, 0x15170000, 0x135d0008, 0x1ed3125c, 0x4f3f5317, 0x00000000, 0x00000000}},
{{17, 0x00216978, 0x22002210, 0x349a0001, 0x1ed31257, 0x4f3f5317, 0x00000000, 0x00000000}},
{{36, 0x19640000, 0x0101000d, 0x10002100, 0x0007a120, 0x196434c0, 0x34a90001, 0x1ed31257}},
{{35, 0x17e50001, 0x1ed31272, 0x4f3f5317, 0x00000000, 0x00000000, 0x00000000, 0x00000123}},
{{12, 0x000000ff, 0x01010000, 0x00000001, 0x00000123, 0x12a917e5, 0x128c0000, 0x53171ed3}},
{{43, 0x04000300, 0x00010000, 0x00010001, 0x1f250000, 0x060402ff, 0x53171ed9, 0x00004f3f}},
{{43, 0x04000300, 0x00000001, 0x00000101, 0x1f2d0100, 0x060402ff, 0x53171ed9, 0x00004f3f}},
{{12, 0xe10000ff, 0x00000000, 0x00ff0002, 0x00000000, 0xe10034c0, 0x203d0000, 0x53171edc}},
{{17, 0x00216978, 0x22002210, 0x204e0001, 0x53171edc, 0x00004f3f, 0x00000000, 0x00000000}},
{{12, 0x19640000, 0x0101000d, 0x10002100, 0x0007a120, 0x196434c0, 0x205d0001, 0x53171edc}},
{{49, 0x01400400, 0x00260300, 0x00880018, 0x00060003, 0x00000006, 0xb6901b00, 0x207c0001}},
{{33, 0x20b40200, 0x208e0031, 0x00000102, 0x53171edf, 0x00004f3f, 0x00000000, 0x00000000}},
{{33, 0x17d50001, 0x20960031, 0x00000102, 0x53171edf, 0x00004f3f, 0x00000000, 0x00000000}},
{{53, 0x01980000, 0x1eec0001, 0x4f3f5317, 0x00000000, 0x00000000, 0x00000000, 0x00000123}},
{{45, 0x00359800, 0x1eefa420, 0x4f3f5317, 0x00000000, 0x00000000, 0x00000000, 0x00000123}},
{{45, 0x00359801, 0x1eefa428, 0x4f3f5317, 0x00000000, 0x00000000, 0x00000000, 0x00000123}},
{{43, 0x04000300, 0x00010100, 0x00010001, 0x1f250001, 0x00000198, 0x53171ef5, 0x00004f3f}},
{{43, 0x04000300, 0x00000101, 0x00000000, 0x1f2d0101, 0x00000198, 0x53171ef5, 0x00004f3f}},
{{35, 0x17dd0100, 0x1efd1886, 0x4f3f5317, 0x00000000, 0x00000000, 0x00000000, 0x00000123}},
{{44, 0x17dd0000, 0x18941886, 0x53171f02, 0x00004f3f, 0x00000000, 0x00000000, 0x01230000}},
{{42, 0x13000300, 0x12c60001, 0x1f05122e, 0x4f3f5317, 0x00000000, 0x00000000, 0x00000000}},
{{4, 0x0f011964, 0x02000100, 0x12db0001, 0x1f05122e, 0x4f3f5317, 0x00000000, 0x00000000}},
{{4, 0x10011964, 0x00000011, 0x12e80001, 0x1f05122e, 0x4f3f5317, 0x00000000, 0x00000000}},
{{50, 0x0f011964, 0x02000100, 0x13130001, 0x122e12eb, 0x53171f05, 0x00004f3f, 0x00000000}},
{{76, 0x01806978, 0x00011401, 0x13200001, 0x122e12eb, 0x53171f05, 0x00004f3f, 0x00000000}},
{{4, 0x08006978, 0x0000ff03, 0x2ddd0001, 0x2f858008, 0x12eb1323, 0x1f05122e, 0x4f3f5317}},
{{78, 0x00003e00, 0x01000091, 0x00010003, 0x2f8a32fd, 0x12eb1323, 0x1f05122e, 0x4f3f5317}},
{{78, 0x00003e00, 0x01000091, 0x00010003, 0x13232f9f, 0x122e12eb, 0x53171f05, 0x00004f3f}},
{{78, 0x00003e00, 0x01000091, 0x00010003, 0x13232fab, 0x122e12eb, 0x53171f05, 0x00004f3f}},
{{78, 0x00003e00, 0x01000091, 0x00010003, 0x13232fbd, 0x122e12eb, 0x53171f05, 0x00004f3f}},
{{4, 0x09003e00, 0x0000ff03, 0x2ddd0001, 0x2fe70009, 0x12eb1323, 0x1f05122e, 0x4f3f5317}},
{{76, 0x0b800000, 0x00010000, 0x320e0001, 0x13232ffd, 0x122e12eb, 0x53171f05, 0x00004f3f}},
{{78, 0x00003e00, 0x01000091, 0x00010003, 0x30083285, 0x12eb1323, 0x1f05122e, 0x4f3f5317}},
{{78, 0x00003e00, 0x01000091, 0x00010003, 0x300b3225, 0x12eb1323, 0x1f05122e, 0x4f3f5317}},
{{78, 0x3e043e00, 0x01000091, 0x00010003, 0x301132b9, 0x12eb1323, 0x1f05122e, 0x4f3f5317}},
{{78, 0x3e343e00, 0x01000091, 0x00010003, 0x302c327b, 0x12eb1323, 0x1f05122e, 0x4f3f5317}},
{{4, 0x0a000000, 0x0000ff03, 0x2ddd0001, 0x3081010a, 0x12eb1323, 0x1f05122e, 0x4f3f5317}},
{{78, 0x00003e00, 0x01000091, 0x00010003, 0x30863285, 0x12eb1323, 0x1f05122e, 0x4f3f5317}},
{{78, 0x00003e00, 0x01000091, 0x00010003, 0x30893225, 0x12eb1323, 0x1f05122e, 0x4f3f5317}},
{{78, 0x3e043e00, 0x01000091, 0x00010003, 0x308f32b9, 0x12eb1323, 0x1f05122e, 0x4f3f5317}},
{{78, 0x3e343e00, 0x01000091, 0x00010003, 0x30aa327b, 0x12eb1323, 0x1f05122e, 0x4f3f5317}},
{{78, 0x00003e00, 0x01000091, 0x00010003, 0x30f23285, 0x12eb1323, 0x1f05122e, 0x4f3f5317}},
{{4, 0x0b003e00, 0x0000ff03, 0x2ddd0001, 0x30f7000b, 0x12eb1323, 0x1f05122e, 0x4f3f5317}},
{{50, 0x01011964, 0x02000100, 0x132d0001, 0x122e12eb, 0x53171f05, 0x00004f3f, 0x00000000}},
{{50, 0x10003105, 0x9fff0032, 0x13350001, 0x122e12eb, 0x53171f05, 0x00004f3f, 0x00000000}},
{{6, 0x17dd0100, 0x16be1886, 0x16a30000, 0x53171f0a, 0x00004f3f, 0x00000000, 0x00000000}},
{{34, 0x17dd0000, 0x16ca1886, 0x1f0a16a3, 0x4f3f5317, 0x00000000, 0x00000000, 0x00000000}},
{{4, 0x0d0003c5, 0x0000ff03, 0x2ddd0001, 0x16e4000d, 0x1f0a16a8, 0x4f3f5317, 0x00000000}},
};

void replay_f(struct atom_context *atom_ctx)
{
{
    unsigned index = GetIndexIntoMasterTable(DATA, VRAM_UsageByFirmware);
    uint8_t frev, crev;
    uint16_t size, offset;
    ATOM_VRAM_USAGE_BY_FIRMWARE_V1_5 *targs;
    ATOM_FIRMWARE_VRAM_RESERVE_INFO_V1_5 *args;

    atom_parse_data_header(atom_ctx, index, &size, &frev, &crev, &offset);
    targs = atom_ctx->bios + offset;
    args = &targs->asFirmwareVramReserveInfo[0];
    printf("VRAM_UsageByFirmware.ulStartAddrUsedByFirmware = 0x%08x\n", args->ulStartAddrUsedByFirmware);
    printf("VRAM_UsageByFirmware.usFirmwareUseInKb = 0x%08x\n", args->usFirmwareUseInKb);
    printf("VRAM_UsageByFirmware.usFBUsedByDrvInKb = 0x%08x\n", args->usFBUsedByDrvInKb);
    printf("ATOM_DP_TRAINING_TBL_ADDR 0x%08x\n", ATOM_DP_TRAINING_TBL_ADDR);
}
}

int main(int argc, char *argv[])
{
    struct ctx ctx;
    unsigned i;
    int r;

    ctx.bios_size = 64 << 10;
    ctx.bios = calloc(1, ctx.bios_size);
    if (ctx.bios == NULL) {
        return -ENOMEM;
    }
    if (radeon_pci_init(&ctx.pradeon)) {
        return -1;
    }
    r = pci_device_read_rom(ctx.pradeon.dev, ctx.bios);
    if (r) {
        goto out;
    }
    r = atombios_init(&ctx);
    if (r) {
        goto out;
    }

#if 1
    fprintf(stderr, "ASIC INIT --------------------------\n");
    r = atom_asic_init(ctx.atom_ctx);
    if (r) {
        fprintf(stderr, "asic init failed\n");
        goto out;
    }
    fprintf(stderr, "ASIC REPLAY ------------------------\n");
    replay_f(ctx.atom_ctx);
#else
    for (i = 0; i < (sizeof(logs)/sizeof(logs[0])); i++) {
        atom_trace_cmd_header(ctx.atom_ctx, logs[i].data[0], &logs[i].data[1]);
    }
#endif


out:
    atombios_fini(&ctx);
    radeon_pci_fini(&ctx.pradeon);
    return r;
}
